use std::path::{Path, PathBuf};

use serde::{Deserialize, Serialize};
use time::{macros::format_description, OffsetDateTime};

use radicle_ci_broker::msg::{Oid, RepoId, RunId, RunResult};

/// Metadata about a run.
#[derive(Debug, Serialize, Deserialize)]
#[serde(deny_unknown_fields)]
#[allow(dead_code)]
pub struct RunInfo {
    /// Repository ID.
    pub repo: String,

    /// Commit ID.
    pub commit: String,

    /// Identifier for the CI run.
    pub id: String,

    /// Result of the CI run.
    pub result: RunResult,

    /// Name of log file.
    pub log: PathBuf,

    /// Name of run info file.
    #[serde(skip)]
    pub run_info: Option<PathBuf>,

    /// Timestamp of when the run ended (the value was created).
    /// ISO8601 format.
    pub timestamp: String,
}

impl RunInfo {
    pub fn builder() -> RunInfoBuilder {
        RunInfoBuilder::default()
    }

    pub fn write(&self) -> Result<(), RunInfoError> {
        if let Some(filename) = &self.run_info {
            let yaml = serde_yaml::to_string(&self).map_err(RunInfoError::SerializeRunInfo)?;
            std::fs::write(filename, yaml.as_bytes())
                .map_err(|e| RunInfoError::WriteRunInfo(filename.into(), e))?;
        }
        Ok(())
    }
}

#[derive(Debug, Default)]
#[allow(dead_code)]
pub struct RunInfoBuilder {
    repo: Option<RepoId>,
    commit: Option<Oid>,
    run_id: Option<String>,
    result: Option<RunResult>,
    log: Option<PathBuf>,
    run_info: Option<PathBuf>,
}

impl RunInfoBuilder {
    pub fn repo(&mut self, repo: RepoId) {
        self.repo = Some(repo);
    }

    pub fn commit(&mut self, commit: Oid) {
        self.commit = Some(commit);
    }

    pub fn run_id(&mut self, id: RunId) {
        self.run_id = Some(format!("{}", id));
    }

    pub fn result(&mut self, result: RunResult) {
        self.result = Some(result);
    }

    pub fn log(&mut self, state: &Path, log: PathBuf) {
        self.log = Some(log.strip_prefix(state).unwrap().into());
    }

    pub fn run_info(&mut self, filename: PathBuf) {
        self.run_info = Some(filename);
    }

    pub fn build(&self) -> Result<RunInfo, RunInfoError> {
        let fmt = format_description!("[year]-[month]-[day] [hour]:[minute]:[second]");
        let now = OffsetDateTime::now_utc()
            .format(fmt)
            .map_err(RunInfoError::TimeFormat)?;

        Ok(RunInfo {
            repo: self.repo.map(|x| x.to_string()).unwrap_or("".into()),
            commit: self.commit.map(|x| x.to_string()).unwrap_or("".into()),
            id: self.run_id.as_deref().unwrap_or("<unknown>").into(),
            result: self
                .result
                .clone()
                .ok_or(RunInfoError::MissingInfo("result".into()))?,
            log: self.log.as_deref().unwrap_or(Path::new("<unknown>")).into(),
            run_info: self.run_info.clone(),
            timestamp: now,
        })
    }
}

#[derive(Debug, thiserror::Error)]
pub enum RunInfoError {
    #[error("programming error: field {0} is not set")]
    MissingInfo(String),

    #[error("failed to write file for run metadata: {0}")]
    WriteRunInfo(PathBuf, #[source] std::io::Error),

    #[error("failed to serialize run metadata: programming error")]
    SerializeRunInfo(#[source] serde_yaml::Error),

    #[error("failed to format current time as a string")]
    TimeFormat(#[source] time::error::Format),
}
